/**
   $Id: Service.java,v 1.17 2006/09/03 02:48:39 orts_mburo Exp $
   $Source: /usr/bodo1/cvs/cvssoftware/ggs/src/ggs/service/common/Service.java,v $
   
   Service class - handles everything

   (c) Michael Buro, licensed under GPL2
*/

package ggs.service.common;

import java.io.*;
import java.net.*;
import java.util.*;

import ggs.shared.*;

// todo: add table destruction code (called when no player left?)
// needs to add id of freed table to free_table_ids

public class Service implements EventHandler,Sender
{
  // inner classes for handling commands
  
  /** command interface */

  abstract class ServiceCmd
  {
    /** return true iff exec OK */
    abstract public void exec(Client client, SReader first);

    /** send error message to client */
    public void error(Client client, String cmd, String usage, String msg)
    {
      sendError(client.getName(), cmd + " - " + msg + ".\\ usage: " + cmd + " " + usage);
    }
  }


  /** element of command list (see below) */
     
  class ServiceCmdData
  {
    public ServiceCmdData(String cmd1, String cmd2, int permission, ServiceCmd exec)
    {
      this.cmd1 = cmd1;
      this.cmd2 = cmd2;    
      this.permission = permission;
      this.exec = exec;
    }
  
    public String     cmd1, cmd2;
    public int        permission;
    public ServiceCmd exec;
  }


  /*
    cmd: create <table-type> [<passwd>]

    create table
  */

  class ExecCreate extends ServiceCmd
  {
    public void exec(Client client, SReader first)
    {
      String tableType = first.nextWord();
      String pw = first.nextWord();

      if (tableType == null) {
	error(client, "table type missing");
	return;
      }

      // create table and check match type

      Table table = tableCreator.newTable();

      if (!table.init(getThis(), getRand(), client.getName(), tableType, pw)) {
	error(client, "illegal table type " + tableType);
	return; 
      }

      int fn = free_table_ids.size();
      int id;
      
      if (fn > 0) {
	// fill hole
	id = free_table_ids.get(fn-1).intValue();
	tables.set(id, table);
	free_table_ids.remove(fn-1);
	
      } else {
	// add element
	id = tables.size();
	tables.add(table);
      }

      table.setId(id);

      // send table creation message to all users who are
      // interested in service messages. We need persistent user
      // flags! Work around: add flags and let client set them
      // when they arrive. Currently the subscriber flag is switched
      // on when sending a list command to the service

      sendToSubscribers("list\\+ " + table.getTableInfo());
      send(client.getName(),
	   "state " + id + " " + table.view(client.getName()));
    }
    
    void error(Client client, String msg)
    {
      super.error(client, "create", "<table-type> [<passwd>]", msg);
    }
  }

  /*
    cmd: table <table-id> ...

    table command
  */

  class ExecTable extends ServiceCmd
  {
    public void exec(Client client, SReader first)
    {
      String tid = first.nextWord();
      if (tid == null) {
	error(client, "table: missing parameter");
	return;
      }
    
      int table_number;

      try { table_number = Integer.parseInt(tid); }
      catch (NumberFormatException e) {
	error(client, tid + " not a table number");
	return;
      }

      if (table_number < 0 || table_number >= tables.size() ||
	  tables.get(table_number) == null) {
	error(client, tid + " not a table number");
	return;
      }
    
      // call table cmd

      Table table = tables.get(table_number);

      if (table.execInService(client.getName(), first, null)) {

	// when cmd sets info change flag send new one to subsribers

	if (table.getInfoChanged()) {
	  sendToSubscribers("list\\+ " + table.getTableInfo());
	}
      } else {
	error(client, "cmd not recognized");
      }
    }

    void error(Client client, String msg)
    {
      super.error(client, "table", "<table-num> command ...", msg);
    }
  }

  /*
    cmd: list

    list command
  */

  class ExecList extends ServiceCmd
  {
    /**
     * Lists all tables that are online
     * 
     */
    public void exec(Client client, SReader first)
    {
      // at first print out number of tables

      if (subscribers.contains(client.getName())) {

	// already subscribing -> no effect

	super.error(client, "list", "", "already subscriber");
	return;
      }

      // sign up for service messages
      subscribers.add(client.getName());

      // send complete list of table infos

      StringBuffer sb = new StringBuffer();
      sb.append("list");
		
      for (Table c : tables) {
	// print out informations about the table
	// FIXME maybe the implizit plus is not correct
	sb.append("\\+ " + c.getTableInfo());
      }

      send(client.getName(), sb.toString());
    }
  }

  /*
   * cmd: join <tn>
   * 
   * join table <tn>
   */

  class ExecJoin extends ServiceCmd
  {
    public void exec(Client client, SReader first)
    {
      String tid = first.nextWord();
      if (tid == null) {
	error(client, "join: missing parameter");
	return;
      }
    
      int table_number;

      try { table_number = Integer.parseInt(tid); }
      catch (NumberFormatException e) {
	error(client, tid + " not a table number");
	return;
      }

      if (table_number < 0 || table_number >= tables.size() ||
	  tables.get(table_number) == null) {
	error(client, tid + " not a table number");
	return;
      }
    
      if (!tables.get(table_number).join(client.getName())) {
	error(client, "you already joined that table");
      }

      send(client.getName(),
	   "state " + tid + " " + tables.get(table_number).view(client.getName()));
    }

    void error(Client client, String msg)
    {
      super.error(client, "join", "<table> join", msg);
    }
  }



  
  /*
    cmd: quit
   
    quit the service (need to be root)
  */

  class ExecQuit extends ServiceCmd
  {
    public void exec(Client client, SReader first)
    {
      Global.shut_down(0);
    }
  }


  /** constructor */
  
  public Service(TableCreator tc, String args[])
  {
    tableCreator = tc;

    Global.options.put("ho", "opal.cs.ualberta.ca"); // host, empty -> keyboard input
    //Global.options.put("ho", ""); // host, empty -> keyboard input
    Global.options.put("po", new Integer(5000));     // port
    Global.options.put("lo", tc.getLogin());
    Global.options.put("pa", "foomoo5");             // fixme : load from config file
    Global.options.put("test", new Integer(0));
    
    // user options (no warnings)

    for (int i=0; i < args.length; i++)
      Global.options.put_line(args[i], false);

    host   = Global.options.getString("ho");
    port   = Global.options.getInteger("po");
    login  = Global.options.getString("lo");
    passwd = Global.options.getString("pa");

    int test = Global.options.getInteger("test");
    if (test != 0) {
      //gameBP.test();
      Global.shut_down(10);
    }
  }

  public void run()
  {
    Misc.msg("starting " + tableCreator.getName());

    if (Global.options.getString("ho").equals("")) {
      // dummy user in local test mode
      name2client.put("root", new Client("root", Client.PERM_ROOT));
    }
    
    // connect to GGS

    running = true;

    sc = new SocketComm();
    
    try { sc.connect(host, port, login, passwd); }

    catch (UnknownHostException f) {
      Global.err("connection problems: unknown host " + host, 2);
    }
    
    catch (IOException e) {
      Global.err("connect problems: io error " + host + ":" + port, 2);
    }

    login = Global.options.getString("lo");

    // login established

    sef = new ServerEventFactory(sc);
    sef.ef.set_default_handler(this);
    sef.start();

    while (running) {

      // fixme: synchronize access!
      checkTime();
      Misc.sleep(100);
    }
  }

  /** execute client command */
  // important: only one thread in check_time and handle_event  
  synchronized public boolean handle_event(EventMsg msg)
  {
    Misc.msg(msg.get_sender());

    for (String s : msg.get_args())
      Misc.msg(s);
    
    if (msg.get_args().size() == 0) {
      Misc.msg("empty input");
      return false;
    }

    SReader r1 = new SReader(msg.get_args().get(0));
    String sender = r1.nextWord();

    Misc.msg("msg: " + msg.get_args().get(0));
    Misc.msg("first: " + sender);
    
    if (sender == null)
      return true;

    sender = sender.substring(0, sender.length()-1);

    // command dispatch

    if (sender.equals("")) {

      // GGS message

      String cmd = r1.nextWord();

      if (cmd == null)
	return true;

      // handle connect/disconnect message

      if (cmd.equals("+")) {

	// list of connected clients
	
	for (String s : msg.get_args()) {

	  // client connected

	  String[] conn = Misc.split(s, "\\s+");	  
	  int perm = 0;
	  try { perm = Integer.parseInt(conn[3]); }
	  catch (NumberFormatException e) {
	    Global.err(conn[3] + " not a permission", 10);
	  }

	  name2client.put(conn[2], new Client(conn[2], perm));

	  Misc.msg("connected: " + conn[2]);
	}
	
      } else if (cmd.equals("-")) {

	// list of disconnected clients
	
	for (String s : msg.get_args()) {

	  // client disconncted

	  String[] conn = Misc.split(s, "\\s+");

	  Client client = name2client.get(conn[2]);
	  
	  if (client == null) {
	    Misc.err("client not in map!");
	  }
	  
	  name2client.remove(conn[2]);

	  for (Table t : tables) {
	    t.disconnected(client.getName());
	  }
	  
	  Misc.msg("disconnected: " + client.getName());
	}
      }

    } else {

      // client message

      Client client = name2client.get(sender);

      if (client == null)
	Misc.err("can't find client with name " + sender);

      // find command
    
      String cmd = r1.nextWord();
      boolean executed = false;

      for (ServiceCmdData e : exec) {

	if (!e.cmd1.equals(cmd) && !e.cmd2.equals(cmd))
	  continue;

	if (client.getPermission() < e.permission)
	  continue;

	// execute it
	e.exec.exec(client, r1);
	executed = true;
	break;
      }

      if (!executed) {
	sendError(client.getName(), "unknown command " + cmd);
	return false;
      }
    }

    return true;
  }

  /** check all games in progress for time violations */
  // important: only one thread in check_time and handle_event
  synchronized private void checkTime()
  {
  }

  public void send(String msg)
  {
    try { sc.write_line(msg); }
    catch (IOException e) {
      Global.err("write error", 11);
    }
  }
  
  public void send(String client, String msg)
  {
    send("t " + client + " " + msg);
  }

  void sendError(String client, String msg)
  {
    send("t " + client + " error: " + msg + "\n");
  }

  void sendToSubscribers(String msg)
  {
    for (String s : subscribers)
      send("t " + s + " " + msg);
  }

  // data

  // false => exit service
  boolean running;

  private String host;
  private int    port;
  private String login;
  private String passwd;  
  
  private SocketComm sc;
  private ServerEventFactory sef;
  private Random rand = new Random();
  
  /** game blueprint for creating games and accessing "static" game members */
  private TableCreator tableCreator;

  /** tables currently served */
  private Vector<Table> tables = new Vector<Table>();
  public Vector<Integer> free_table_ids = new Vector<Integer>(); // for recycling table ids

  /** maps names to connected clients, is owner of clients */
  private HashMap<String, Client> name2client = new HashMap<String, Client>();

  /** list of users who listen to table messages (initiated by sending
   *  list cmd). May be replaced by persistent client flags stored in file later.
   */
  private Set<String> subscribers = new TreeSet<String>();
  
  private ServiceCmdData exec[] = {
    new ServiceCmdData("quit"   , "q" , Client.PERM_SUPER_USER, new ExecQuit()   ),
    new ServiceCmdData("create" , "c" , Client.PERM_UNREG_USER, new ExecCreate() ),
    new ServiceCmdData("table"  , "t" , Client.PERM_UNREG_USER, new ExecTable()  ),    
    new ServiceCmdData("list"   , "l" , Client.PERM_UNREG_USER, new ExecList()   ),
    new ServiceCmdData("join"   , "j" , Client.PERM_UNREG_USER, new ExecJoin()   )      
  };

  public Random getRand() { return rand; }
  private Service getThis() { return this; }
}
