/** 
    $Id: SkatState.java,v 1.10 2006/09/04 03:40:42 orts_mburo Exp $
    $Source: /usr/bodo1/cvs/cvssoftware/ggs/src/ggs/service/skat/SkatState.java,v $

    Stand-alone skat game module

    (c) Michael Buro, licensed under GPL2
*/

package ggs.service.skat;

import java.io.*;
import java.util.*;
import java.text.*;
import ggs.service.common.*;
import ggs.shared.*;


public class SkatState
{
  static public int FORE_HAND   = 0;
  static public int MIDDLE_HAND = 1;
  static public int REAR_HAND   = 2;  

  enum Phase { BIDDING, HAND, DISCARD, CARDPLAY, FINISHED };
  
  public Hand[] hands;
  public Vector<Card> skat;    // skat (or discarded cards)
  public Phase phase;          // game phase
  public int to_move;          // starts with 1 (first bidder), 0 plays first card
  public int max_bid;          // highest bid so far
  public boolean bid_expected; // otherwise answer is expected
  public int bidder;           // bidding player
  public int asked;            // by the bidder
  public int declarer;         // who won the bidding (-1: not declared yet)
  public GameDeclaration decl; // declared game type
  public Vector<Card> trick;   // (partial) trick
  public int trick_winner;     // so far
  public Card winning_card;    // so far
  public int[] points;         // total points so far for each player
  public int[] trick_num;      // won by each player

  // client mode variables
  public int clientType;              // who am I? (>=0:player, <0:observer)
  public boolean[] clientPerspective; // whom am I observing?
  
  /**
     read position from string
     @return null if correct, error message otherwise
  */
  public String setStartState(String s) { Misc.err("setstartstate not impl."); return "foo"; }

  /**
     write position to string
  */
  public String getStartStateView(int view)
  {
    StringBuffer sb = new StringBuffer();
    int sum = 0;
    for (int i=0; i < 3; ++i) {
      if (view == StdTable.VIEW_ALL || i == view) {
	sum++;
	sb.append(cards_to_string(hands[i].cards));
      } else {
	sb.append("?");
      }
      if (i < 2)
	sb.append(DEL);
    }

    // no skat

    return sb.toString();
  }

  /**
     @return bit set that contains 1 for each player to move
  */
  public long playersToMove()
  {
    if (finished())
      return 0;
    return 1 << to_move;
  }

  /**
     @return true iff game is finished
  */
  public boolean finished()
  {
    return phase == Phase.FINISHED;
  }

  /**
     @return game result for a player
  */
  public int result(int player) { return 0; }

  /**
     @return maximum result in position POS for a player
  */
  public int maxResult(int player) { return 0; }

  /**
     @return null if game type is OK, error message otherwise
     ("static function")
  */
  public String checkGameType(String gt, int[] player_nums)
  {
    player_nums[0] = player_nums[1] = 3;
    return null;
  }

  /**
     init state w.r.t. game type and number of players
     @param game_type  type of game
     @param player_num number of players
     @param gen        random number generator
     @return null if parameters correct, error message otherwise
  */
  public String initState(String game_type, int player_num, Random gen)
  {
    if (player_num != 3) 
      return "need exactly three players";

    // create random position
    
    phase = Phase.BIDDING;
    bid_expected = true;
    bidder = MIDDLE_HAND;
    asked = FORE_HAND; // match deals with rotation
    max_bid = 17; // < 18 indicates no bid yet
    declarer = -1;
    to_move = bidder;
    
    // generate 32 card deck

    Vector<Card> deck = new Vector<Card>();

    for (int suit=0; suit < 4; ++suit) {
      for (int kind=0; kind < 8; ++kind) {
	deck.add(new Card(suit, kind));
      }
    }

    // shuffle deck

    Collections.shuffle(deck, gen);

    if (false) { // old code
      for (int i=0; i < 31; ++i) {
	// swap random element at position >= i with element at i
	int j = gen.nextInt(32-i)+i;
	Card t = deck.get(i);
	deck.set(i, deck.get(j));
	deck.set(j, t);
      }
    }

    // deal cards
    hands  = new Hand[3];

    for (int i=0; i < 3; ++i) {
      hands[i] = new Hand();
      for (int j=0; j < 10; ++j) {
	hands[i].cards.add(deck.get(10*i+j));
      }
    }
    
    skat = new Vector<Card>();
    skat.add(deck.get(30));
    skat.add(deck.get(31));

    points    = new int[3]; // set to 0
    trick_num = new int[3]; // set to 0    

    trick = new Vector<Card>();

    return null;
  }

  // last move data
  
  int    lastPlayer;
  String lastMove;

  String lastPublicMove;      // what is shown to observers
  String lastPrivateMove;     // what is shown to player and kibitzers
  String lastPublicWorldMove; // what is shown to observers
  String lastPrivateWorldMove;// what is shown to player and kibitzers  
  String lastWorldMoveData;   // data required to reproduce world move (e.g. random...)

  /**
   * @param user perspective (bit i true iff player i watched)
   * @return view of last move
   */
  public String lastMoveData()
  {
    return lastMove + (lastWorldMoveData != null ? ":" + lastWorldMoveData : "");
  }

  /**
   * @param user perspective (bit i true iff player i watched)
   * @return view of last move
   */
  public String lastMoveView(int perspective)
  {
    if (perspective == StdTable.VIEW_ALL || perspective == lastPlayer) {
      // player or kibitz
      return lastPrivateMove + " " + lastPrivateWorldMove;
    }

    // public or other player
    return lastPublicMove + " " + lastPublicWorldMove;
  }

  /**
   * make move
   * @param player who moves
   * @param move to be played
   * @param doit change state, if false only check validity
   * @param inClient true iff move executed in client
   * @return error message if move is illegal, otherwise execute move
   * and create player move views iff doit is true and return null
   */
  public String makeMove(int player, String move, boolean doit, boolean inClient)
  {
    if (player != to_move)
      return "not your turn";

    lastPlayer = player;
    lastMove = move;
    lastWorldMoveData = null;
    
    lastPublicMove = lastPrivateMove = move; // default public = private = move
    lastPublicWorldMove = lastPrivateWorldMove = "";
    
    switch (phase) {

    case BIDDING:

      if (bid_expected) {

	if (move.equals("pass")) {

	  // pass

	  // move OK

	  if (!doit)
	    return null;

	  if (bidder == FORE_HAND) {

	    // no bids: game cancelled

	    phase = Phase.FINISHED;
	    
	  } else if (bidder == MIDDLE_HAND) {

	    bidder = REAR_HAND;
	    asked = FORE_HAND;
	    to_move = bidder;
	    
	  } else {

	    if (max_bid < 18) {

	      // forehand hasn't bid anything
	      // so ask whether he wants to play
	      bidder = FORE_HAND;
	      asked = FORE_HAND;
	      to_move = bidder;

	    } else {

	      // otherwise asked player becomes declarer
	      phase = Phase.HAND;
	      declarer = asked;
	      to_move = declarer;
	    }
	  }

	} else {

	  // move must be number > last bid

	  int bid;
	  try { bid = Integer.parseInt(move); }
	  catch (NumberFormatException e) {
	    return move + " not a bid";
	  }

	  if (bid <= max_bid)
	    return "bid must be higher than " + max_bid;

	  if (bid > 500)
	    return "ridiculous bid " + move;

	  // move OK
	  
	  if (!doit)
	    return null;

	  max_bid = bid;
	  bid_expected = false; // answer expected
	  to_move = asked;

	  if (player == FORE_HAND) { // forehand bids, others passed
	    phase = Phase.HAND;
	    declarer = FORE_HAND;
	    to_move = declarer;
	  }

	}
	
      } else {

	// answer yes/no expected

	if (move.equals("yes")) {

	  // move OK
	  
	  if (!doit)
	    return null;

	  // continue bidding
	  
	  bid_expected = true;
	  to_move = bidder;

	} else if (move.equals("pass")) {

	  // move OK
	  
	  if (!doit)
	    return null;

	  if (bidder == MIDDLE_HAND) {

	    // start second bidding
	    
	    bid_expected = true;
	    bidder = REAR_HAND;
	    asked  = MIDDLE_HAND;
	    to_move = bidder;
	    
	  } else {

	    phase = Phase.HAND;
	    declarer = REAR_HAND;
	    to_move = declarer;

	  }

	} else {

	  return "expected yes|pass, got " + move;	  
	  
	}
      }

      break;

      // ----------------------------------------
      
    case HAND:

      // move: skat | game declaration

      if (move.equals("skat")) {

	// move OK
	
	if (!doit)
	  return null;

        // also send skat to declarer
        lastPrivateWorldMove = skat.get(0).toString() + SEQ_DEL + skat.get(1).toString();
 	phase = Phase.DISCARD;
	to_move = declarer;

      } else {

	GameDeclaration new_decl = new GameDeclaration();
	String msg = new_decl.fromString(move);
	if (msg != null) {
	  return "skat | game-decl expected, got " + move + ": " + msg;
	}

	// move OK
	
	if (!doit)
	  return null;

	decl = new_decl;
	phase = Phase.CARDPLAY;
	to_move = FORE_HAND;
      }
      
      break;

      // ----------------------------------------      

    case DISCARD:

      // move: card1,card2,game declaration

      String[] parts = Misc.split(move, SEQ_DEL);
      if (parts.length != 3)
	return "card1,card2,game-decl expected, got " + move;

      Card c1 = new Card();
      String msg = c1.fromString(parts[0]);
      if (msg != null)
	return "expected card1,card2,game-decl, got " + move + ": illegal 1st card";

      Card c2 = new Card();
      msg = c2.fromString(parts[1]);
      if (msg != null)
	return "expected card1,card2,game-decl, got " + move + ": illegal 2nd card";
      
      GameDeclaration new_decl = new GameDeclaration();
      msg = new_decl.fromString(parts[2]);
      if (msg != null)
	return "expected card1,card2,game-decl, got " + move + ": illegal game-decl";

      // syntax OK

      // check whether c1 and c2 are in players hand + skat

      Set<Card> all = new TreeSet<Card>(new GreaterCard());

      all.addAll(hands[declarer].cards);
      all.add(skat.get(0));
      all.add(skat.get(1));      

      if (false) {
	// test
	for (Card c : all)
	  System.out.print(c.toString() + " ");
	System.out.print(c1.toString() + " " + c2.toString());
      }
      
      if (!all.contains(c1))
	return "you don't hold " + parts[0];
      
      if (!all.contains(c2))
	return "you don't hold " + parts[1];	

      // discarded cards OK

      // todo: check whether game-decl can fit the bid, do we care?

      // move OK
      
      if (!doit)
	return null;

      // don't send discarded cards to opponents
      lastPublicMove = parts[2];

      // update current information
      
      skat.set(0, c1);
      skat.set(1, c2);
      all.remove(c1);
      all.remove(c2);

      hands[declarer].cards.clear();
      for (Card c : all)
	hands[declarer].cards.add(c);

      decl = new_decl;      

      phase = Phase.CARDPLAY;
      to_move = FORE_HAND;
      break;

      // ----------------------------------------

    case CARDPLAY:

      // move: card

      Card c = new Card();
      msg = c.fromString(move);
      if (msg != null)
	return msg;

      // check whether player has card

      if (!hands[to_move].cards.contains(c))
	return "you don't have " + move;

      if (trick.size() > 0 && !card_ok(c))
	return "you must follow suit";

      // move ok

      if (!doit)
	return null;

      trick.add(c);
      hands[to_move].cards.remove(c);
      
      if (trick.size() == 1) {
	trick_winner = to_move;
	winning_card = c;
	to_move = (to_move + 1) % 3;
	return null;
      }
      
      if (highest_card(c)) {
	trick_winner = to_move;
	winning_card = c;
      }
      
      if (trick.size() == 2) {

	to_move = (to_move + 1) % 3;

      } else {

	if (hands[to_move].cards.size() == 0)
	  phase = Phase.FINISHED;

	trick_num[trick_winner]++;
	points[trick_winner] += trick_points();

	to_move = trick_winner;
	trick.clear();
      }
      
      break;
    }

    return null;
  }
  
  // -------------------------------------------------
  // helper functions
  // -------------------------------------------------

  private static String DEL = ":";     // top-level delimiter
  private static String SEQ_DEL = ","; // sequence delimiter
  
  /*
   *  @return true iff card OK in current trick
   */
  private boolean card_ok(Card c)
  {
    if (trick.size() == 0)
      return true;

    if (follows_suit(c))
      return true;

    if (trump(c))
      return true;

    return !has_trump(to_move);
  }

  /**
     @return true iff card currently wins trick
  */
  private boolean highest_card(Card c)
  {
    if (trick.size() == 0)
      return true;

    if (trump(c)) {

      if (!trump(winning_card))
	return true;

      return rank(c) > rank(winning_card);

    } else {

      if (trump(winning_card))
	return false;
      
      if (!follows_suit(c))
	return false;

      return rank(c) > rank(winning_card);
    }
  }

  
  /**
     @return return numerical card rank
  */
  public int rank(Card c)
  {
    if (decl.type == GameDeclaration.Type.NULL)
      return Card.null_ranks[c.rank]; // 10 and J are low

    if (trump(c)) {

      if (c.rank == Card.RANK_JACK)
	return 7 + c.suit;

      return c.rank;
    }

    return c.rank; // no trump
  }


  /**
     @return true iff card follows suit in current trick
  */
  private boolean follows_suit(Card c)
  {
    if (trick.size() == 0)
      return true;

    if (trump(c))
      return trump(trick.get(0));

    return c.suit == trick.get(0).suit;
  }

  
  /**
     @return true iff card is trump
  */
  private boolean trump(Card c)
  {
    if (decl.type == GameDeclaration.Type.NULL)
      return false;

    if (decl.type == GameDeclaration.Type.GRAND)
      return c.rank == Card.RANK_JACK;

    // suit games

    if (c.rank == Card.RANK_JACK)
      return true;
      
    return c.suit == decl.trump_suit();
  }


  /**
     @return true iff player has trump cards
  */
  private boolean has_trump(int player)
  {
    if (decl.type == GameDeclaration.Type.NULL)
      return false;

    if (decl.type == GameDeclaration.Type.GRAND) {

      for (Card c : hands[player].cards) {
	if (c.rank == Card.RANK_JACK)
	  return true;
      }

      return false;
    }

    // suit game

    int trump_suit = decl.trump_suit();
      
    for (Card c : hands[player].cards) {

      if (c.rank == Card.RANK_JACK)
	return true;

      if (c.suit == trump_suit)
	return true;
    }
    
    return false;
  }


  /**
     @return points in completed trick
  */
  private int trick_points()
  {
    if (trick.size() != 3)
      Global.err("trick not completed", 30);

    int points = 0;
    for (Card c : trick) {
      points += Card.points[c.rank];
    }
    return points;
  }

  /** convert card vector to string, ? if empty */
  static private String cards_to_string(Collection<Card> cards)
  {
    if (cards.size() == 0)
      return "?";

    StringBuffer s = new StringBuffer();
    boolean first = true;

    for (Card c : cards) {
      if (!first) {
	s.append(SEQ_DEL);
      }
      first = false;
      s.append(c);
    }

    return s.toString();
  }


  // ------------ client functions -------------
  
  /** convert string to card vector, ? : empty
      @return null if error in string, card vector otherwise
  */
  static public Vector<Card> string_to_cards(String s)
  {
    Vector<Card> cards = new Vector<Card>();

    if (!s.equals("?")) {
      String[] card_strs = s.split(SEQ_DEL, 0);
      for (String cs : card_strs) {
 	Card c = new Card();
 	if (c.fromString(cs) != null)
 	  throw new NumberFormatException();
 	cards.add(c);
      }
    }

    return cards;
  }

  static public Phase string_to_phase(String s)
  {
    if (s.length() != 1)
      throw new NumberFormatException();

    switch (s.charAt(0)) {
    case 'B': return Phase.BIDDING;
    case 'H': return Phase.HAND;
    case 'D': return Phase.DISCARD;
    case 'C': return Phase.CARDPLAY;
    default:
      throw new NumberFormatException();      
    }
  }

  static public String phase_to_string(Phase p)
  {
    switch (p) {
    case BIDDING:  return "B";
    case HAND:	   return "H";
    case DISCARD:  return "D";  
    case CARDPLAY: return "C";  
    }
    return "";
  }

//   /**
//      @param user >=0: player: <0: observer
//      @return game state in readable form
//   */
//   public String historyView(int user, boolean[] perspective)
//   {
//     boolean[] show = new boolean[3];
//     StringBuffer sb = new StringBuffer();

//     sb.append(user
// 	      + DEL + phase_to_string(phase)	      
// 	      + DEL + to_move
// 	      + DEL + max_bid
// 	      + DEL + (bid_expected?1:0)
// 	      + DEL + bidder
// 	      + DEL + asked
// 	      + DEL + declarer
// 	      + DEL + (decl == null ? "?" : decl)
// 	      + DEL + cards_to_string(trick)
// 	      );

//     // hands
    
//     // show hand if perspective bit is set
//     for (int i = 0; i < 3; ++i)
//       show[i] = perspective[i];

//     // or if open game was declared
//     if (decl != null && decl.ouvert)
//       show[declarer] = true;

//     for (int i = 0; i < 3; ++i) {
//       if (show[i])
// 	sb.append(DEL + cards_to_string(hands[i].cards));
//       else
// 	sb.append(DEL + "?");
//     }

//     // skat only visible if game declared and declarer is observed

//     if (declarer >= 0 && perspective[declarer])
//       sb.append(DEL + cards_to_string(skat));
//     else 
//       sb.append(DEL + "?");

//     sb.append(DEL);
    
//     // perspective (0,0,1,0,1...)
//     for (int i=0; i < perspective.length; ++i) {
//       sb.append(perspective[i] ? 1 : 0);
//       if (i < perspective.length-1)
// 	sb.append(SEQ_DEL);
//     }
    
//     return sb.toString();
//   }

//   /**
//      set game state from string
//      @return null if error, error message otherwise
//   */
//   public String fromView(String s)
//   {
//     String[] vals = s.split(DEL, 0);

//     if (vals.length != 15)
//       return "view string corrupt (length) : " + s;

//     int i = 0;

//     try {
//       clientType    = Integer.parseInt(vals[i]);      i++;
//       phase         = string_to_phase(vals[i]);       i++;
//       to_move       = Integer.parseInt(vals[i]);      i++;
//       max_bid       = Integer.parseInt(vals[i]);      i++;
//       bid_expected  = Integer.parseInt(vals[i]) != 0; i++;      
//       bidder        = Integer.parseInt(vals[i]);      i++;
//       asked         = Integer.parseInt(vals[i]);      i++;
//       declarer      = Integer.parseInt(vals[i]);      i++;

//       // game declaration
//       if (vals[i].equals("?"))
//  	decl = null;
//       else {
//  	decl = new GameDeclaration();	
//  	if (decl.fromString(vals[i]) != null)
//  	  throw new NumberFormatException();
//       }
//       i++;

//       // trick
//       trick = string_to_cards(vals[i]); i++;
  
//       // skat
//       skat = string_to_cards(vals[i]); i++;

//       // hands
//       for (int j=0; j < 3; ++j) {
// 	hands[i].cards = string_to_cards(vals[i++]);
//       }

//       // perspective

//       String [] vals2 = s.split(SEQ_DEL, 0);
//       clientPerspective = new boolean[vals2.length];
//       for (int j=0; j < vals2.length; ++j)
// 	clientPerspective[j] = !vals[j].equals("0");
//     }

//     catch (NumberFormatException e) {
//       return "illegal value (" + i + ")";
//     }

//     return null;
//   }

}


class Card 
{
  Card() { }
  
  Card(int suit, int rank) 
  {
    this.suit = suit;
    this.rank = rank;
  }

  public int suit; // 0:DIAMONDS 1:HEARTS 2:SPADES 3:CLUBS
  public int rank; // 0:7 ... 5:TEN 6:ACE 7:JACK

  private static String[] suit_names = { "D", "H", "S", "C" }; // ordered by value
  private static String[] rank_names = { "7", "8", "9", "Q", "K", "T", "A", "J" };

  public static int[]     null_ranks = {  0,   1,   2,   5,   6,   3,   7,   4 };
  public static int[]     points     = {  0,   0,   0,   3,   4,  10,  11,   2 };  

  public static int SUIT_DIAMONDS = 0;
  public static int SUIT_HEARTS   = 1;
  public static int SUIT_SPADES   = 2;
  public static int SUIT_CLUBS    = 3;

  public static int RANK_JACK     = 7;  
  
  public String toString()
  {
    return suit_names[suit] + rank_names[rank];
  }

  public boolean equals(Object o)
  {
    if (o == null || !(o instanceof Card))
      return false;

    Card co = (Card)o;
    
    return suit == co.suit && rank == co.rank;
  }

  /**
   *  @return null if correct, error message otherwise
   */
  public String fromString(String s)
  {
    if (s.length() != 2)
      return "card: two characters expected, got " + s;

    // read suit
    
    String c1 = s.substring(0, 1);
    for (suit=0; suit < 4; ++suit)
      if (suit_names[suit].equals(c1))
	break;

    if (suit >= 4)
      return "card: suit character not recognized, got " + s;

    // read rank
    
    String c2 = s.substring(1, 2);
    for (rank=0; rank < 8; ++rank)
      if (rank_names[rank].equals(c2))
	break;

    if (rank >= 8)
      return "card: rank character not recognized, got " + s;

    return null;
  }

}

class GreaterCard implements Comparator<Card>
{
  public int compare(Card first, Card second)
  {
    int r = first.suit - second.suit;
    if (r != 0)
      return r;
    return first.rank - second.rank;
  }
}



class Hand
{
  public Vector<Card> cards = new Vector<Card>();
}
 

class GameDeclaration
{
  enum Type { DIAMONDS, HEARTS, SPADES, CLUBS, GRAND, NULL };

  public Type    type = Type.NULL;
  public boolean hand;
  public boolean ouvert;
  public boolean schneider_announced;
  public boolean schwarz_announced;

  /**
     @return index of trump suit (in Card order)
   */
  public int trump_suit()
  {
    switch (type) {

    case DIAMONDS: return 0;
    case HEARTS:   return 1;
    case SPADES:   return 2;
    case CLUBS:    return 3;

    default:       return -1;
    }
  }

  
  public String toString()
  {
    StringBuffer s = new StringBuffer();
    
    switch (type) {

    case DIAMONDS: s.append("D"); break;
    case HEARTS:   s.append("H"); break;
    case SPADES:   s.append("S"); break;
    case CLUBS:    s.append("C"); break;
    case GRAND:    s.append("G"); break;      
    case NULL:     s.append("N"); break;

    }

    if (ouvert)
      s.append("O");
    
    if (hand)
      s.append("H");

    if (schneider_announced)
      s.append("!");

    if (schwarz_announced)
      s.append("#");

    return s.toString();
  }

  /**
     @return null if correct, error message otherwise
   */
  public String fromString(String s)
  {
    StringCharacterIterator sci = new StringCharacterIterator(s);
    char c = sci.first();

    ouvert = hand = schneider_announced = schwarz_announced = false;
    
    if (c == CharacterIterator.DONE) return "empty";

    switch (c) {

    case 'D': type = Type.DIAMONDS; break;
    case 'H': type = Type.HEARTS; break;
    case 'S': type = Type.SPADES; break;
    case 'C': type = Type.CLUBS; break;
    case 'G': type = Type.GRAND; break;      
    case 'N': type = Type.NULL; break;

    default: return "type '" + c + "' not recognized";
    }

    for (;;) {
      c = sci.next();
      if (c == CharacterIterator.DONE) break;

      switch (c) {
      case 'O': ouvert = true; continue;
      case 'H': hand = true; continue;
      case '!': schneider_announced = true; continue;
      case '#': schwarz_announced = true; continue;
      default: return "class '" + c + "' not recognized";
      }
    }

    if (schwarz_announced)
      schneider_announced = true;

    // check exceptions
    
    if (!hand && schneider_announced)
      return "not hand but schneider or schwarz announced";
    
    if (type == Type.NULL) {
      if (schneider_announced || schwarz_announced)
	return "null game but schneider or schwarz announced";

    } else {

      if (ouvert) {
	hand = true;
	schneider_announced = schwarz_announced = true;
      }
    }

    return null;
  }
  
}

