/* $Id: Options.java,v 1.1 2006/08/29 09:10:42 orts_mburo Exp $
   
   (c) Michael Buro, licensed under the GPL2

   Options class

   Allows to map strings to values of different types and
   to set overwrite default values by command line arguments

   The latest addition is indirection which is used for
   default parameters

   example:

   // definition:
   
   options.put("port", new Integer(5000));

   // reading args:

   options.put_line(arg_string, true);

   // using option:

   Integer port = options.getInteger("port");


   // Indirection

   options.put("def.foo", new Integer(5)); // default
   
   options.puti("os.foo", "def.foo"); // indirection

   if os.foo is not set in put_line, get("os.foo") refers
   to value of def.foo. Otherwise, a new value with type
   of def.foo is created and used for os.foo.
   
*/

package ggs.shared;

import java.util.*;

public class Options
{
  static final char escape      = '\\';
  static final char arg_sep     = '%';
  static final char space_repl  = '_';
  static final char new_prefix  = '.';

  static final int STR   = 1;
  static final int INT   = 2;
  static final int GEO   = 3;
  static final int BOOL  = 4;
  static final int DOUBLE = 5;  
  static final int INDIR  = 6; // indirection to other option
  
  private Hashtable<String, Opt> ht = new Hashtable<String, Opt>();

  private abstract class Opt
  {
    boolean orig = true;
    int type = 0;
    abstract void fromString(String s) throws IllegalArgumentException;
    abstract Opt copy();
  }
  
  private class OptBool extends Opt
  {
    Boolean value;

    OptBool(Boolean v) { value = v; type = BOOL; }

    void fromString(String s)
    {
      if (s.equals("0")) 
	value = Boolean.FALSE;
      else
	value = Boolean.TRUE;
    }

    public OptBool copy() { return new OptBool(value); }

    public String toString() { return value.toString(); }
  }

  private class OptInt extends Opt
  {
    Integer value;

    OptInt(Integer v) { value = v; type = INT; }
    
    void fromString(String s) { value = Integer.decode(s); }

    public OptInt copy() { return new OptInt(value); }

    public String toString() { return value.toString(); }    
  }

  private class OptStr extends Opt
  {
    String value;

    OptStr(String v) { value = v; type = STR; }

    void fromString(String s) { value = s; }

    public OptStr copy() { return new OptStr(value); }

    public String toString() { return value.toString(); }    
  }

  private class OptDouble extends Opt
  {
    Double value;

    OptDouble(Double v) { value = v; type = DOUBLE; }

    void fromString(String s) { value = Double.valueOf(s); }

    public OptDouble copy() { return new OptDouble(value); }

    public String toString() { return value.toString(); }
  }

  private class OptGeo extends Opt
  {
    Geometry value;

    OptGeo(Geometry v) { value = v; type = GEO; }

    void fromString(String s) { value = Geometry.valueOf(s); }

    public OptGeo copy() { return new OptGeo(value); }

    public String toString() { return value.toString(); }
  }

  /** indirection */
  private class OptInd extends Opt
  {
    String value;

    OptInd(String v) { value = v; type = INDIR; }

    void fromString(String s) { value = s; }

    public OptInd copy() { return new OptInd(value); }

    public String toString() { return "[ " + value + " ]"; }
  }


  public Options() { }

  /** check whether key is already present */
  private void check(String key)
  {
    if (ht.get(key) != null)
      Misc.err("options: multiple definition of " + key);
  }

  public void put(String key, Integer v)
  {
    check(key);
    put2(key, v);
  }

  // overwrite  
  public void put2(String key, Integer v)
  {
    ht.put(key, new OptInt(v));
  }

  public void put(String key, Double v)
  {
    check(key);
    put2(key, v);
  }

  // overwrite  
  public void put2(String key, Double v)
  {
    ht.put(key, new OptDouble(v));
  }

  public void put(String key, Boolean v)
  {
    check(key);
    put2(key, v);
  }

  // overwrite  
  public void put2(String key, Boolean v)
  {
    ht.put(key, new OptBool(v));
  }

  public void put(String key, String v)
  {
    check(key);
    put2(key, v);
  }

  // overwrite  
  public void put2(String key, String v)
  {
    ht.put(key, new OptStr(v));
  }

  public void put(String key, Geometry v)
  {
    check(key);
    put2(key, v);
  }

  // overwrite  
  public void put2(String key, Geometry v)
  {
    ht.put(key, new OptGeo(v));
  }

  
  public void puti(String key, String v)
  {
    check(key);
    put2i(key, v);
  }

  // overwrite  
  public void put2i(String key, String v)
  {
    ht.put(key, new OptInd(v));
  }


  // get value (possibly after one step of indirection)
  private Opt getv(String key)
  {
    Opt o = ht.get(key);

    if (o == null)
      return null;
    
    if (o instanceof OptInd) {
      // Global.dbgmsg("indirection " + key + " -> " + ((OptInd)o).value);
      return ht.get(((OptInd)o).value); // indirection
    }
    return o;
  }
  
  public Boolean getBoolean(String key)
  {
    OptBool opt = (OptBool)getv(key);
    if (opt == null || opt.type != BOOL) return null;
    return opt.value;
  }

  public Integer getInteger(String key)
  {
    OptInt opt = (OptInt)getv(key);
    if (opt == null || opt.type != INT) return null;
    return opt.value;
  }

  public Double getDouble(String key)
  {
    OptDouble opt = (OptDouble)getv(key);
    if (opt == null || opt.type != DOUBLE)
      return null;
    return opt.value;
  }

  public String getString(String key)
  {
    OptStr opt = (OptStr)getv(key);
    if (opt == null || opt.type != STR) return null;
    return opt.value;
  }

  public Geometry getGeometry(String key)
  {
    OptGeo opt = (OptGeo)getv(key);
    if (opt == null || opt.type != GEO) return null;
    return opt.value;
  }

  public boolean put_line(String line0, boolean warnings)
  {
    // parse option line
    // % separates options, observe \%
    // _ -> space, observe \_
    // .w  -> w is prefix for following options

    // Global.dbgmsg("opt: "+line0);
    
    String line = "";

    // delete '\n's

    for (int i=0; i < line0.length(); i++) {
      if (line0.charAt(i) != '\n') line += line0.charAt(i);
    }

    line += arg_sep;
    
    String prefix = "";
    int l = line.length();
    String opt = "";

    for (int i=0; i < l; i++) {

      char c = line.charAt(i);

      if (c == escape && i+1 < l) {
	if (line.charAt(i+1) == space_repl || line.charAt(i+1) == arg_sep) {
	  opt += line.charAt(i+1);
	  i++;
	  continue;
	}
      } 

      if (c == space_repl) c = ' ';

      if (c == arg_sep) {   // next parameter

	if (opt.length() > 0 && opt.charAt(0) == new_prefix) {

	  // new prefix
	  
	  prefix = opt.substring(1);

	} else {

	  // new option
	  // first part is key, second part is value

	  int j = opt.indexOf(' ');
	  String key, val;
	  
	  if (j < 0) {
	    key = prefix + opt;
	    val = "";
	  } else {
	    key = prefix + opt.substring(0, j);
	    val = opt.substring(j+1, opt.length());
	  }

	  Opt hopt = ht.get(key);
	  
	  if (hopt == null) {

	    // option not recognized

	    if (warnings) {
	      Misc.msg("WARNING: opt=\""+key+"\" [val=\""+val+"\"] not recognized");
	    }
	    opt = "";
	    continue;
	  }


	  if (hopt instanceof OptInd) {

	    // indirection: create copy of pointed to variable to be overwritten

	    OptInd o = (OptInd)hopt;
	    hopt = ht.get(o.value);

	    if (hopt == null) {
	      if (warnings) {
		Misc.msg("WARNING: indirection not found "+key+" -> "+o.value);
	      }
	      opt = "";
	      continue;
	    }

	    hopt = hopt.copy();
	    ht.put(key, hopt);
	  }

	  Misc.msg("set opt \""+key+"\" to \""+val+"\"");

	  try { hopt.fromString(val); }
	  catch (NumberFormatException e) {
	    Misc.msg("Illegal option value");
	  }
	  
	  opt = "";
	}

	continue;
      }
      
      opt += c;
    }

    return true;
  }
  

  public void write()
  {
    Enumeration en = ht.keys();

    while (en.hasMoreElements()) {
      String key = (String) en.nextElement();
      
      Misc.msg(key + " : " + ht.get(key).toString());
    }
  }
}

